@@ -1,6 +1,13 @@
-import json
+import os
+import io
+import ujson as json
+import uvloop
 import asyncio
 
+from types import MappingProxyType
+
+from uuid import uuid4
+
 import aiohttp
 
 from aiohttp import ClientResponse
@@ -9,7 +16,7 @@ from aiologger import Logger
 from aiologger.levels import LogLevel
 from aiologger.formatters.base import Formatter
 
-from typing import Optional, Dict, List, Union
+from typing import Optional, Dict, List, Union, Coroutine
 
 from threading import Thread
 
@@ -19,6 +26,22 @@ from .helpers import InlineKeyboardMarkup, Format
 from .middleware import BaseBotMiddleware
 
 
+def read_file(filepath: str) -> io.BytesIO:
+
+    assert os.path.exists(filepath), f'File "{filepath}" does\'nt exist'
+    assert os.path.isfile(filepath), f'"{filepath}" is not a file'
+
+    with open(filepath, 'rb') as f:
+        file_obj = io.BytesIO(f.read())
+        file_obj.name = os.path.basename(filepath)
+        return file_obj
+
+
+async def async_read_file(filepath: str):
+    loop = asyncio.get_event_loop()
+    return await loop.run_in_executor(None, read_file, filepath)
+
+
 def keyboard_to_json(
         keyboard_markup: Union[
             List[List[Dict]], InlineKeyboardMarkup, str, None
@@ -53,18 +76,42 @@ def format_to_json(format_: Union[Format, List[Dict], str]):
 
 class AsyncBot(object):
 
+    __slots__ = (
+        "loop",
+        "url",
+        "session",
+        "base_url",
+        "parseMode",
+        "token",
+        "proxy",
+        "logger",
+        "running",
+        "handlers",
+        "middlewares",
+        "lastEventId",
+        "pollTime",
+        "__polling_thread"
+    )
+
     def __init__(
             self,
             token: str,
             url: str = 'https://myteam.mail.ru',
             parseMode: str = 'HTML',
             proxy: Optional[str] = None,
-            middlewares: List[BaseBotMiddleware] = [],
+            log_level: LogLevel = LogLevel.INFO,
+            middlewares: List[BaseBotMiddleware] = (),
             lastEventId: int = 0,
-            pollTime: int = 30
+            pollTime: int = 30,
+            loop: Optional = None
     ):
 
-        self.loop = asyncio.new_event_loop()
+        # gc.disable()
+        if loop is None:
+            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
+            self.loop = asyncio.new_event_loop()
+        else:
+            self.loop = loop
 
         self.session = None
         self.url: str = url
@@ -82,7 +129,7 @@ class AsyncBot(object):
                 '%(module)s:%(funcName)s:%(lineno)d - '
                 '%(message)s'
             ),
-            level=LogLevel.INFO
+            level=log_level
         )
 
         self.running = True
@@ -97,18 +144,38 @@ class AsyncBot(object):
         self.loop.run_until_complete(self.start_session())
 
         BaseBotMiddleware.bot = self
+        Event.bot = self
+
+        # gc.freeze()
+        # gc.enable()
+
+    @staticmethod
+    def get_request_id() -> str:
+        """
+        Метод для создания уникального uuid запроса
+        :return: уникальный uuid запроса
+        """
+        return str(uuid4())
+
+    @staticmethod
+    def loads(object) -> MappingProxyType:
+
+        return MappingProxyType(
+            json.loads(object)
+        )
 
     async def start_session(self):
         """
         Функция создания асинхронной сессии
         :return: None
         """
+
         self.session: aiohttp.ClientSession = aiohttp.ClientSession(
             base_url=self.url,
             raise_for_status=True,
-            read_timeout=40,
-            conn_timeout=4,
-
+            timeout=aiohttp.ClientTimeout(total=self.pollTime + 5),
+            json_serialize=json.dumps,
+            loop=asyncio.get_event_loop()
         )
 
     async def get(self, path: str, **kwargs) -> ClientResponse:
@@ -118,6 +185,9 @@ class AsyncBot(object):
         :param kwargs: параметры GET-запроса
         :return: ответ сервера
         """
+
+        request_id = self.get_request_id()
+
         params = {
             'token': self.token
         }
@@ -130,7 +200,7 @@ class AsyncBot(object):
                 params.pop(key)
 
         await self.logger.debug(
-            f'[GET] /bot/v1/{path} params - {kwargs} ->'
+            f'[GET][{request_id}] /bot/v1/{path} params - {kwargs} ->'
         )
 
         response = await self.session.get(
@@ -139,36 +209,51 @@ class AsyncBot(object):
             proxy=self.proxy
         )
         await self.logger.debug(
-            f'<- [{response.status}] /bot/v1/{path}'
+            f'<- [{response.status}][{request_id}] /bot/v1/{path}'
         )
         return response
 
-    async def post(self, path: str, **kwargs) -> ClientResponse:
+    async def post(
+            self,
+            path: str,
+            data: Union[Dict[str, str], Dict[str, io.BytesIO]] = None,
+            **kwargs
+    ) -> ClientResponse:
         """
         Функция для создания и логирования POST-запроса
         :param path: относительный path запроса
+        :param data:
         :param kwargs: параметры POST-запроса
         :return: ответ сервера
         """
+
+        request_id = self.get_request_id()
+
         params = {
             'token': self.token
         }
 
         params.update(kwargs)
 
+        for key, value in params.copy().items():
+
+            if value is None:
+                params.pop(key)
+
         self.logger.debug(
-            f'[POST] {path} kwargs - {kwargs} ->'
+            f'[POST][{request_id}] {path} kwargs - {kwargs} ->'
         )
 
-        async with self.session.post(
-                url=f'{self.base_url}{path}',
+        response = await self.session.post(
+                url=f'/bot/v1/{path}',
                 params=params,
+                data=data,
                 proxy=self.proxy
-        ) as response:
-            self.logger.debug(
-                f'<- [{response.status}]'
-            )
-            return await response.json()
+        )
+        self.logger.debug(
+            f'<- [{response.status}][{request_id}]'
+        )
+        return response
 
     async def self_get(self,) -> ClientResponse:
         """
@@ -192,9 +277,9 @@ class AsyncBot(object):
             self,
             chatId: str,
             text: str,
-            replyMsgId: Optional[List[int]] = None,
+            replyMsgId: Optional[List[str]] = None,
             forwardChatId: Optional[str] = None,
-            forwardMsgId: Optional[List[int]] = None,
+            forwardMsgId: Optional[List[str]] = None,
             inlineKeyboardMarkup: Union[
                 List[List[Dict[str, str]]], InlineKeyboardMarkup, None] = None,
             _format: Union[Format, List[Dict], str, None] = None,
@@ -301,6 +386,41 @@ class AsyncBot(object):
             parseMode=parseMode if parseMode is not None else self.parseMode
         )
 
+    async def send_file(
+        self,
+        chatId: str,
+        file_path: str,
+        caption: Optional[str] = None,
+        replyMsgId: Optional[List[int]] = None,
+        forwardChatId: Optional[str] = None,
+        forwardMsgId: Optional[List[int]] = None,
+        inlineKeyboardMarkup: Union[
+                List[List[Dict[str, str]]], InlineKeyboardMarkup, None] = None,
+        _format: Union[Format, List[Dict], str, None] = None,
+        parseMode: Optional[str] = None
+    ) -> ClientResponse:
+        """
+        Метод для отправки сообщения с файлом по его file.
+        """
+        return await self.post(
+            path='messages/sendFile',
+            chatId=chatId,
+            data={'file': await async_read_file(file_path)},
+            caption=caption,
+            replyMsgId=replyMsgId,
+            forwardChatId=forwardChatId,
+            forwardMsgId=forwardMsgId,
+            inlineKeyboardMarkup=inlineKeyboardMarkup
+            if isinstance(inlineKeyboardMarkup, str)
+            else keyboard_to_json(inlineKeyboardMarkup),
+            format=_format
+            if isinstance(_format, str)
+            else format_to_json(_format),
+            parseMode=parseMode
+            if parseMode is not None
+            else self.parseMode
+        )
+
     async def send_voiceId(
             self,
             chatId: str,
@@ -347,6 +467,42 @@ class AsyncBot(object):
             else keyboard_to_json(inlineKeyboardMarkup),
         )
 
+    async def send_voice(
+        self,
+        chatId: str,
+        file_path: str,
+        caption: Optional[str] = None,
+        replyMsgId: Optional[List[int]] = None,
+        forwardChatId: Optional[str] = None,
+        forwardMsgId: Optional[List[int]] = None,
+        inlineKeyboardMarkup: Union[
+                List[List[Dict[str, str]]], InlineKeyboardMarkup, None] = None,
+        _format: Union[Format, List[Dict], str, None] = None,
+        parseMode: Optional[str] = None
+    ) -> ClientResponse:
+        """
+        Метод для отправки сообщения с голосового сообщения по его file,
+        он должен быть в формате aac, ogg или m4a.
+        """
+        return await self.post(
+            path='messages/sendVoice',
+            chatId=chatId,
+            data={'file': await async_read_file(file_path)},
+            caption=caption,
+            replyMsgId=replyMsgId,
+            forwardChatId=forwardChatId,
+            forwardMsgId=forwardMsgId,
+            inlineKeyboardMarkup=inlineKeyboardMarkup
+            if isinstance(inlineKeyboardMarkup, str)
+            else keyboard_to_json(inlineKeyboardMarkup),
+            format=_format
+            if isinstance(_format, str)
+            else format_to_json(_format),
+            parseMode=parseMode
+            if parseMode is not None
+            else self.parseMode
+        )
+
     async def edit_text(
             self,
             chatId: str,
@@ -621,7 +777,13 @@ class AsyncBot(object):
             chatId: str,
             cursor: Optional[str] = None
     ) -> ClientResponse:
-
+        """
+        Метод для получения списка пользователей чата
+        :param chatId: ID чата
+        :param cursor: курсор, используется если количество пользователей более
+         определенного лимита
+        :return: Список пользователей чата
+        """
         return await self.get(
             path="chats/getAdmins",
             chatId=chatId,
@@ -632,7 +794,11 @@ class AsyncBot(object):
             self,
             chatId: str,
     ) -> ClientResponse:
-
+        """
+        Метод для получения списка заблокированных пользователей в чате
+        :param chatId: ID чата
+        :return: Список заблокированных пользователей
+        """
         return await self.get(
             path="chats/getBlockedUsers",
             chatId=chatId,
@@ -642,7 +808,11 @@ class AsyncBot(object):
             self,
             chatId: str,
     ) -> ClientResponse:
-
+        """
+        Метод для получения списка пользователей, желающих вступить в чат
+        :param chatId: ID чата
+        :return: список пользователей
+        """
         return await self.get(
             path="chats/getPendingUsers",
             chatId=chatId,
@@ -654,7 +824,13 @@ class AsyncBot(object):
             userId: str,
             delLastMessages: bool = True
     ) -> ClientResponse:
-
+        """
+        Метод для блокировки пользователей в чате
+        :param chatId: ID чата
+        :param userId: ID пользователя
+        :param delLastMessages: удалять ли сообщения пользователя
+        :return: результат запроса
+        """
         return await self.get(
             path="chats/blockUser",
             chatId=chatId,
@@ -667,7 +843,12 @@ class AsyncBot(object):
             chatId: str,
             userId: str,
     ) -> ClientResponse:
-
+        """
+        Метод для разблокировки пользователя в чате
+        :param chatId: ID чата
+        :param userId: ID пользователя
+        :return: результат запроса
+        """
         return await self.get(
             path="chats/unblockUser",
             chatId=chatId,
@@ -681,7 +862,15 @@ class AsyncBot(object):
             userId: Optional[str] = None,
             everyone: bool = True
     ) -> ClientResponse:
-
+        """
+        Метод для одобрения вступления пользователя в чат
+        :param chatId: ID чата
+        :param approve: разрешить или запретить в вступление
+        :param userId: ID пользователя
+        :param everyone: Флаг для разрешения вступления всех пользователей что
+        хотят в чат вступить
+        :return: результат запроса
+        """
         if everyone:
             return await self.get(
                 path="chats/resolvePending",
@@ -703,7 +892,12 @@ class AsyncBot(object):
             chatId: str,
             title: str
     ) -> ClientResponse:
-
+        """
+        Метод для установки названия чата
+        :param chatId: ID чата
+        :param title: новое название
+        :return: результат запроса
+        """
         return await self.get(
             path="chats/setTitle",
             chatId=chatId,
@@ -715,7 +909,12 @@ class AsyncBot(object):
             chatId: str,
             about: str
     ) -> ClientResponse:
-
+        """
+        Метод для установки описания чата
+        :param chatId: ID чата
+        :param about: новое описание чата
+        :return: результат запроса
+        """
         return await self.get(
             path="chats/setAbout",
             chatId=chatId,
@@ -727,7 +926,12 @@ class AsyncBot(object):
             chatId: str,
             rules: str
     ) -> ClientResponse:
-
+        """
+        Метод для установки новых правил чата
+        :param chatId: ID чата
+        :param rules: новые правила чата
+        :return: результат запроса
+        """
         return await self.get(
             path="chats/setRules",
             chatId=chatId,
@@ -737,45 +941,83 @@ class AsyncBot(object):
     async def pin_msg(
             self,
             chatId: str,
-            msgId: int
+            msgId: str
     ) -> ClientResponse:
-
+        """
+        Метод для закрепления сообщения в чате
+        :param chatId: ID чата
+        :param msgId: ID сообщения
+        :return: результат запроса
+        """
         return await self.get(
             path="chats/pinMessage",
             chatId=chatId,
             msgId=msgId,
         )
 
+    async def unpin_msg(
+            self,
+            chatId: str,
+            msgId: str
+    ) -> ClientResponse:
+        """
+        Метод для закрепления сообщения в чате
+        :param chatId: ID чата
+        :param msgId: ID сообщения
+        :return: результат запроса
+        """
+        return await self.get(
+            path="chats/unpinMessage",
+            chatId=chatId,
+            msgId=msgId,
+        )
+
     async def get_file_info(
             self,
             fileId: str
     ) -> ClientResponse:
-
+        """
+        Получение информации о файле по его fileId
+        :param fileId: ID файла
+        :return: информация о файле
+        """
         return await self.get(
             path="files/getInfo",
             fileId=fileId
         )
 
-    async def get_events(self):
+    @staticmethod
+    def loads(object, *args, **kwargs) -> MappingProxyType:
 
+        return MappingProxyType(
+            json.loads(object)
+        )
+
+    async def get_events(self):
+        """
+        Метод для поллинга событий от Bit API
+        :return: Список событий
+        """
         response = await self.get(
             path="events/get",
             lastEventId=self.lastEventId,
             pollTime=self.pollTime
         )
 
-        response_json = await response.json()
+        response_json = await response.json(
+            loads=self.loads
+        )
 
         if response_json.get('events', []):
 
             self.lastEventId = response_json['events'][-1]['eventId']
 
-        return response_json
+        return response_json['events']
 
     async def handle_wrapper(self, handler, event: Event):
 
         try:
-            await handler(self, event)
+            await handler(event)
         except Exception as error:
             await self.logger.exception(error)
 
@@ -793,68 +1035,61 @@ class AsyncBot(object):
                     result = await middleware.check(event_)
                     if result:
                         return True
-                else:
-                    if middleware.check(event_):
-                        return True
+                elif middleware.check(event_):
+                    return True
         return False
 
-    async def start_polling(self):
+    def task_check(self, event_, handler, event_type, cmd) -> Optional[Coroutine]:
+        if (event_type == event_.type and cmd is None) \
+                or \
+                (event_type == EventType.NEW_MESSAGE == event_.type \
+                 and event_.text.startswith(cmd)):
 
+            return self.handle_wrapper(
+                handler=handler,
+                event=event_
+            )
+
+    async def start_polling(self):
+        """
+        Функция поллинга и обработки событий
+        :return:
+        """
         while self.running:
 
-            try:
-                events = await self.get_events()
-
-                events = [
-                    Event(
+            for event_ in map(
+                    lambda event: Event(
                         type_=EventType(event["type"]),
                         data=event["payload"]
-                    )
-                    for event in events['events']
-                ]
+                    ),
+                    await self.get_events()
+            ):
+                await self.logger.debug(event_)
 
-                tasks = []
+                if await self.middleware_check(event_):
+                    continue
 
-                for event_ in events:
-                    await self.logger.debug(event_)
-
-                    middleware_check = await self.middleware_check(event_)
-                    if middleware_check:
-                        continue
-
-                    for handler, event_type, cmd in self.handlers:
-                        if event_type == event_.type and cmd is None:
-                            tasks.append(
-                                self.handle_wrapper(
-                                    handler=handler,
-                                    event=event_
-                                )
-                            )
-                        elif cmd is not None \
-                                and event_type == EventType.NEW_MESSAGE \
-                                and event_type == event_.type \
-                                and event_.text.startswith(cmd):
-                            tasks.append(
-                                self.handle_wrapper(
-                                    handler=handler,
-                                    event=event_
-                                )
-                            )
-                        else:
-                            await self.logger.debug(
-                                f'Passing event: {event_.type} {event_.data}'
-                            )
-
-                if tasks:
-                    await asyncio.wait(tasks, timeout=0)
-            except (
-                    asyncio.exceptions.TimeoutError,
-                    aiohttp.client_exceptions.ServerTimeoutError
-            ) as error:
-                await self.logger.error(error)
+                await asyncio.wait(
+                    filter(
+                        lambda x: x is not None,
+                        map(
+                            lambda x: self.task_check(
+                                event_, *x,
+                            ),
+                            self.handlers
+                        )
+                    ),
+                    timeout=0
+                )
 
     def start_poll(self, threaded: bool = False):
+        """
+        Перегрузка функции поллинга и обрабоки событий
+        :param threaded: выполнять функцию в основном потоке
 
+        или запустить побочный
+        :return:
+        """
         if threaded:
 
             self.__polling_thread = Thread(
@@ -869,7 +1104,11 @@ class AsyncBot(object):
             self.loop.run_until_complete(self.start_polling())
 
     def add_handler(self, handler: List):
-
+        """
+        Функция добавления обработчика событий
+        :param handler:
+        :return:
+        """
         if asyncio.iscoroutinefunction(handler[0]):
             self.handlers.append(handler)
         else:
@@ -878,7 +1117,12 @@ class AsyncBot(object):
             )
 
     def event_handler(self, event_type: EventType, cmd: Optional[str] = None):
-
+        """
+        Базовый декоратор для функции обработки события
+        :param event_type: ALL
+        :param cmd: ?
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -903,6 +1147,7 @@ class AsyncBot(object):
         """
         Декоратор для функций обработки входящих сообщений
         :param event_type: тип события = EventType.NEW_MESSAGE
+        :param cmd:
         :return:
         """
         def decorate(handler):
@@ -918,6 +1163,7 @@ class AsyncBot(object):
         """
         Декоратор для функций обработки входящих сообщений
         :param event_type: тип события = EventType.NEW_MESSAGE
+        :param cmd:
         :return:
         """
         def decorate(handler):
@@ -930,7 +1176,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.EDITED_MESSAGE,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события редактирования сообщения
+        :param event_type: тип события = EventType.EDIT_MESSAGE
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -941,7 +1192,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.DELETED_MESSAGE,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события удаления сообщения
+        :param event_type: тип события = EventType.DELETE_MESSAGE
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -952,7 +1208,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.PINNED_MESSAGE,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события закрепления сообщения в чате
+        :param event_type: тип события = EventType.PIN_MESSAGE
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -963,7 +1224,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.UNPINNED_MESSAGE,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события открепления сообщения в чате
+        :param event_type: тип события = EventType.UNPIN_MESSAGE
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -974,7 +1240,13 @@ class AsyncBot(object):
             event_type: EventType = EventType.NEW_CHAT_MEMBERS,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для фукнции обработки
+        события добавления нового пользователя в чат
+        :param event_type: тип события = EventType.NEW_CHAT_MEMBERS
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -985,7 +1257,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.LEFT_CHAT_MEMBERS,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события удаления пользователя из чата
+        :param event_type: тип события = EventType.LEFT_CHAT_MEMBERS
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -996,7 +1273,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.CHANGED_CHAT_INFO,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события обновления информации о чате
+        :param event_type: тип события = EventType.CHANGED_CHAT_INFO
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
@@ -1007,7 +1289,12 @@ class AsyncBot(object):
             event_type: EventType = EventType.CALLBACK_QUERY,
             cmd: Optional[str] = None
     ):
-
+        """
+        Декоратор для функции обработки события клика по кнопке с коллбеком
+        :param event_type: тип события = EventType.CALLBACK_QUERY
+        :param cmd: -
+        :return:
+        """
         def decorate(handler):
             self.add_handler([handler, event_type, cmd])
             return handler
